#include <stdio.h>
#include <strings.h>

#include "sage++user.h"

#include "anls_att.h"

extern SgType * gettype(int);

extern SgExpression * find_exp(int);
extern void tree_exp(SgExpression *, int);

extern char * mixstring;
extern Flag in_main_program;
extern Flag in_procedure;

extern char * stackNameOfType(SgType *);
extern char * nameOfType(SgType *);
extern char * sizeNameOfType(SgType *);

void add_include(SgStatement * s)
{
  SgStatement * firstdecl=s->childList1(0);
  
  SgValueExp * filename = new SgValueExp("pcrc.inc");

  SgStatement * include = new SgStatement(INCLUDE_STAT, NULL, NULL, filename, NULL, NULL);

  firstdecl->insertStmtBefore(*include, *s);
}


void add_init_finalize(SgStatement * s, SgStatement * lastdec, SgStatement * lastexec)
{
  int idsize;

  {
    SgSymbol * fname = s->GetSymbolInScope(lastdec->controlParent(), mixstring, "init", NULL);
    SgStatement * afuncall = new SgCallStmt(*fname);
    lastdec->insertStmtAfter(*afuncall, * lastdec->controlParent());
  };

  {
    SgSymbol * fname = s->GetSymbolInScope(lastdec->controlParent(), mixstring, "finalize", NULL);
    SgStatement * afuncall = new SgCallStmt(*fname);
    lastexec->insertStmtBefore(*afuncall, * lastexec->controlParent());
  };

};

void add_stack_operation(SgStatement * s, SgStatement * lastdec, SgStatement * lastexec)
{
  SgProcHedrStmt * prochedr = (SgProcHedrStmt *) s;

  for (int i = 0; i<prochedr->numberOfParameters(); i++) { 
    SgSymbol * parametersymb = prochedr->parameter(i);

    if (parametersymb->numberOfAttributes(Inherit_Def)!=0) {

      SgType * parametertype = parametersymb->type();
      if (parametertype->variant()==T_ARRAY) {

	SgArrayType * ptype = (SgArrayType *) parametertype;

	for (int j = 0; j<ptype->dimension(); j++) {
	  {
	    SgSymbol * fname = s->GetSymbolInScope(lastdec->controlParent(), mixstring, "reset_push_bounds", NULL);
	    SgCallStmt * afuncall = new SgCallStmt(*fname);

	    SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", parametersymb->identifier());
	    SgExpression * dadexp = new SgExpression ( VAR_REF, NULL, NULL, dad);
	    afuncall->addArg(* dadexp);

	    SgValueExp * dim = new SgValueExp(j+1);
	    afuncall->addArg(* dim);
	    
	    SgTripletExp * size = (SgTripletExp * ) ptype->sizeInDim(j);
	    SgValueExp * start;

	    if (size->numberOfItems()!=0)
	      start = new SgValueExp (size->item(0)->valueInteger());
	    else
	      start = new SgValueExp (1);
	    afuncall->addArg(* start);
	    
	    //SgValueExp * end = new SgValueExp (size->item(1)->valueInteger());
	    //afuncall->addArg(* end);

	    SgSymbol * actlb = s->getSymbolInCurrentScope(mixstring, "alb", parametersymb->identifier(), j);
	    SgExpression * actlbexp = new SgExpression (VAR_REF, NULL, NULL, actlb);

	    afuncall->addArg( *actlbexp);

	    SgSymbol * actub = s->getSymbolInCurrentScope(mixstring, "aub", parametersymb->identifier(), j);
	    SgExpression * actubexp = new SgExpression (VAR_REF, NULL, NULL, actub);

	    afuncall->addArg( *actubexp);

	    SgSymbol * actstr = s->getSymbolInCurrentScope(mixstring, "ast", parametersymb->identifier(), j);
	    SgExpression * actstrexp = new SgExpression (VAR_REF, NULL, NULL, actstr);

	    afuncall->addArg( *actstrexp);

	    lastdec->insertStmtAfter(*afuncall, * lastdec->controlParent());
	  };

	  {
	    SgSymbol * fname = s->GetSymbolInScope(lastdec->controlParent(), mixstring, "reset_pop_bounds", NULL);
	    SgCallStmt * afuncall = new SgCallStmt(*fname);
	    
	    SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", parametersymb->identifier());
	    SgExpression * dadexp = new SgExpression ( VAR_REF, NULL, NULL, dad);
	    afuncall->addArg(* dadexp);

	    SgValueExp * dim = new SgValueExp(j+1);
	    afuncall->addArg(* dim);


	    SgSymbol * actlb = s->getSymbolInCurrentScope(mixstring, "alb", parametersymb->identifier(), j);
	    SgExpression * actlbexp = new SgExpression (VAR_REF, NULL, NULL, actlb);

	    afuncall->addArg( *actlbexp);

	    SgSymbol * actub = s->getSymbolInCurrentScope(mixstring, "aub", parametersymb->identifier(), j);
	    SgExpression * actubexp = new SgExpression (VAR_REF, NULL, NULL, actub);

	    afuncall->addArg( *actubexp);

	    SgSymbol * actstr = s->getSymbolInCurrentScope(mixstring, "ast", parametersymb->identifier(), j);
	    SgExpression * actstrexp = new SgExpression (VAR_REF, NULL, NULL, actstr);

	    afuncall->addArg( *actstrexp);
	
	    lastexec->insertStmtBefore(*afuncall, * lastexec->controlParent());
	  };
	}
      }
    }
  }
};

// this function get the i-th dim of the index;
SgExpression * getithdim(int i, SgExpression * index)
{
  if (i==0) {
    return(index->operand(1));
  }
  else {
    SgExpression * tail = index->operand(2);
    for (int j = 1; j < i; j++) {
      tail = tail->operand(2);
    }
    return (tail->operand(1));
  };
};

void dist_align_array(SgStatement * s, SgStatement * lastdec, SgStatement * lastexec)
{
  for (int i=0; i<s->numberOfAttributes(Align_Def); i++) {
    Alignment_def * adef = (Alignment_def *) ((s->getAttribute(i, Align_Def))->getAttributeData());
    
    //get the arrary length expression;
    SgExpression * arraylength = adef->getfromsymbol()->type()->length();

    //for each dim, add set_array_rang_align;

    for (int j=adef->getDim()-1; j>=0; j--)  {
      SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "set_array_align", NULL);
      SgCallStmt * afuncall = new SgCallStmt(*fname);

      SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", adef->getfromsymbol()->identifier());
      SgExpression * expr = new SgExpression(EXPR_LIST, NULL, NULL, dad);

      afuncall->addArg(*expr);

      SgValueExp * dim = new SgValueExp(j+1);

      afuncall->addArg(*dim);

      SgExpression * jthdim = getithdim(j, arraylength);
      //printf("The %dth dim id is: %d\n", j, jthdim->id());

      SgValueExp * start;

      if (jthdim->operand(1)->operand(1))

	/*--------------
        if (expr->operand(1)->isInteger())
          start=expr->operand(1)->valueInteger();
        else if (expr->operand(1)->symbol()->variant()==CONST_NAME) { // should be a parameter value ;
          //printf ("The id is %d\n", ((SgConstantSymb *) expr->operand(2)->symbol())->constantValue()->id() );
          if (((SgConstantSymb *) expr->operand(1)->symbol())->constantValue()->isInteger())
            start =((SgConstantSymb *) expr->operand(2)->symbol())->constantValue()->valueInteger();
          else
            printf("It should be an integer\n");
        }
        else
          printf("Unknow template boundaries.\n");
	  ------------*/

	start = new SgValueExp(jthdim->operand(1)->operand(1)->valueInteger());
      else
	start = new SgValueExp(1);
	
      afuncall->addArg(*start);
      
      SgValueExp * end = new SgValueExp(jthdim->operand(1)->operand(2)->valueInteger());
      afuncall->addArg(*end);

      SgValueExp * stride = new SgValueExp(adef->getStride(j));
      afuncall->addArg(*stride);

      SgValueExp * offset = new SgValueExp(adef->getOffset(j));
      afuncall->addArg(*offset);      
     
      SgSymbol * tmpsymb = adef->gettosymbol();
      Template_def * tdef = (Template_def *)tmpsymb->getAttribute(0, Temp_Def)->getAttributeData();
      SgSymbol * rngsymb = tdef->getrng();
      
      SgValueExp * dimention = new SgValueExp(j+1);
      SgExpression * range = new SgExpression(RANGE_OP, NULL, dimention, NULL);
      SgExpression * ddot = new SgExpression(DDOT, range, NULL, NULL);
      SgExpression * index = new SgExpression(EXPR_LIST, ddot, NULL, NULL);
      SgExpression * rngarray = new SgExpression(ARRAY_REF, index, NULL, rngsymb);
 
      afuncall->addArg(*rngarray);
      
      lastdec->insertStmtAfter(*afuncall, *s);
    }

    {
      SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "new_array_data", NULL);
      SgFunctionRefExp * funcexpr = new SgFunctionRefExp (*fname);
      
      SgSymbol * array = adef->getfromsymbol();
      SgExpression * arrayexpr = new SgExpression(VAR_REF, NULL, NULL, array);
      funcexpr->addArg(*arrayexpr);

      SgType * stype= adef->getfromsymbol()->type();

      SgSymbol * typeName=s->getSymbolInCurrentScope(mixstring, nameOfType(stype), NULL);
      SgExpression * typeExp = new SgExpression (VAR_REF, NULL, NULL, typeName);
      funcexpr->addArg(*typeExp);
    
      SgSymbol * sizeName = s->getSymbolInCurrentScope(mixstring, sizeNameOfType(stype), NULL);
      SgExpression * sizeExp = new SgExpression (VAR_REF, NULL, NULL, sizeName);
      funcexpr->addArg(*sizeExp);

      SgValueExp * arraydim = new SgValueExp(adef->getDim());
      funcexpr->addArg(*arraydim);
      
      //SgValueExp * dataorder= new SgValueExp(COLUMN_MAJOR);
      //funcexpr->addArg(*dataorder);
      
      // in the following we want the grp name of the prcossor;

      SgSymbol * templatename= adef->gettosymbol();
      Distribute_def * ddef = (Distribute_def *) templatename->getAttribute(0, Dist_Def)->getAttributeData();

      SgSymbol * procname = ddef->gettosymbol();
      Processor_def * pdef = (Processor_def *)procname->getAttribute(0, Proc_Def)->getAttributeData();

      SgSymbol * procgrd = pdef->getgrp();
      SgExpression * procgrdexpr = new SgExpression(VAR_REF, NULL, NULL, procgrd);
      funcexpr->addArg(*procgrdexpr);

      SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", adef->getfromsymbol()->identifier());
      SgExpression * dadexpr = new SgExpression(VAR_REF, NULL, NULL, dad);
      
      SgAssignStmt * assign = new SgAssignStmt(*dadexpr, *funcexpr);

      lastdec->insertStmtAfter(*assign, *s);

    }

    // below is the couter part of the above set_array_data_done;

    {
      SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "delete_array", NULL);
      SgCallStmt * afuncall = new SgCallStmt(*fname);

      SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", adef->getfromsymbol()->identifier());

      SgExpression * expr = new SgExpression(EXPR_LIST, NULL, NULL, dad);

      afuncall->addArg(*expr);

      lastexec->insertStmtBefore(*afuncall, *s);
    }
  };
};

void define_noalign_data(SgStatement * s, SgStatement * lastdec, SgStatement * lastexec)
{
  SgSymbol * first = ((SgProgHedrStmt *)s)->firstSymbolInScope();

  for (; first; first=first->nextInScope()) {
    if (first->numberOfAttributes(Noalign_Def)!=0) {
      SgSymbol * ndsymb= first;
      {
	for (int i=0; i<s->numberOfAttributes(Proc_Def); i++) {

	  Processor_def * pdef = (Processor_def *) ((s->getAttribute(i, Proc_Def))->getAttributeData()); 

	  SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "new_array_data", NULL);
	  SgFunctionRefExp * funcexpr = new SgFunctionRefExp (*fname);
	  
	  SgExpression * symbexp = new SgExpression (VAR_REF, NULL, NULL, ndsymb);
	  funcexpr->addArg(*symbexp);
	  
	  SgType * stype= ndsymb->type();

	  SgSymbol * real=s->getSymbolInCurrentScope(mixstring, nameOfType(stype), NULL);
	  SgExpression * realexp = new SgExpression (VAR_REF, NULL, NULL, real);
	  funcexpr->addArg(*realexp);
      
	  SgSymbol * sreal = s->getSymbolInCurrentScope(mixstring, sizeNameOfType(stype), NULL);
	  SgExpression * srealexp = new SgExpression (VAR_REF, NULL, NULL, sreal);
	  funcexpr->addArg(*srealexp);

	  SgSymbol * grp = s->getSymbolInCurrentScope(mixstring, "grp", pdef->getsymbol()->identifier());
	  SgExpression * grpexp = new SgExpression (VAR_REF, NULL, NULL, grp);
	  funcexpr->addArg(*grpexp);
	  
	  SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", ndsymb->identifier());
	  SgExpression * dadexp = new SgExpression (VAR_REF, NULL, NULL, dad);
	  
	  SgAssignStmt * assign = new SgAssignStmt(*dadexp, *funcexpr);
	  
	  lastdec->insertStmtAfter(*assign, *s);
	}
	
      }
    };
  };

};

void add_template(SgStatement * s, SgStatement * lastdec, SgStatement * lastexec)
{
  
  for (int i=0; i<s->numberOfAttributes(Temp_Def); i++) {
    
    Template_def * tdef = (Template_def *) ((s->getAttribute(i, Temp_Def))->getAttributeData());
    SgSymbol * templatename= tdef->getsymbol();    
    SgSymbol * rngsymb = tdef->getrng();
    Distribute_def * ddef = (Distribute_def *) templatename->getAttribute(0, Dist_Def)->getAttributeData();
    SgSymbol * procname = ddef->gettosymbol();
    Processor_def * pdef = (Processor_def *)procname->getAttribute(0, Proc_Def)->getAttributeData();

    {
      for (int j=tdef->getDim()-1; j>=0; j--) {
	switch (ddef->distCode(j)) {
	case NOP:
	  {
	    SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "new_range_collapse", NULL);
      
	    SgValueExp * dimention = new SgValueExp(j+1);
	    SgExpression * range = new SgExpression(RANGE_OP, NULL, dimention, NULL);
	    SgExpression * ddot = new SgExpression(DDOT, range, NULL, NULL);
	    SgExpression * index = new SgExpression(EXPR_LIST, ddot, NULL, NULL);

	    SgExpression * rngarray = new SgExpression(ARRAY_REF, index, NULL, rngsymb);
	    
	    SgFunctionRefExp * funcexpr = new SgFunctionRefExp (*fname);
	    
	    SgValueExp * start= new SgValueExp(tdef->getStart(j));
	    funcexpr->addArg(*start);
	    
	    SgValueExp * end = new SgValueExp(tdef->getEnd(j));
	    funcexpr->addArg(*end);

	    SgAssignStmt * assign = new SgAssignStmt(*rngarray, *funcexpr);
	    
	    lastdec->insertStmtAfter(*assign, *s);
	    break;
	  }
	case CYCLIC:
	  {
	    SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "new_range_distribute", NULL);
      
	    SgValueExp * dimention = new SgValueExp(j+1);
	    SgExpression * range = new SgExpression(RANGE_OP, NULL, dimention, NULL);
	    SgExpression * ddot = new SgExpression(DDOT, range, NULL, NULL);
	    SgExpression * index = new SgExpression(EXPR_LIST, ddot, NULL, NULL);

	    SgExpression * rngarray = new SgExpression(ARRAY_REF, index, NULL, rngsymb);
	    
	    SgFunctionRefExp * funcexpr = new SgFunctionRefExp (*fname);
	    
	    SgValueExp * start= new SgValueExp(tdef->getStart(j));
	    funcexpr->addArg(*start);
	    
	    SgValueExp * end = new SgValueExp(tdef->getEnd(j));
	    funcexpr->addArg(*end);

	    SgSymbol * procname = ddef->gettosymbol();

	    // distributing code;
	    SgValueExp * dist = new SgValueExp(2);
	    funcexpr->addArg(*dist);

	    // in the following we want the grp name of the prcossor;

	    SgSymbol * procgrd = pdef->getgrp();
	    SgExpression * procgrdexpr = new SgExpression(VAR_REF, NULL, NULL, procgrd);
	    funcexpr->addArg(*procgrdexpr);

	    // we need the processor dim here;

	    int processorDimCounter=0;
	    for (int k=0; k<=j; k++) {
	      if (ddef->distCode(k)!=NOP)
		processorDimCounter++;
	    }
	    SgValueExp * dim = new SgValueExp(processorDimCounter);	      
	    
	    funcexpr->addArg(*dim);
	    
	    SgAssignStmt * assign = new SgAssignStmt(*rngarray, *funcexpr);
	    
	    lastdec->insertStmtAfter(*assign, *s);
	    break;
	  }
	case BLOCK: 
	  {
	    SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "new_range_distribute", NULL);
      
	    SgValueExp * dimention = new SgValueExp(j+1);
	    SgExpression * range = new SgExpression(RANGE_OP, NULL, dimention, NULL);
	    SgExpression * ddot = new SgExpression(DDOT, range, NULL, NULL);
	    SgExpression * index = new SgExpression(EXPR_LIST, ddot, NULL, NULL);

	    SgExpression * rngarray = new SgExpression(ARRAY_REF, index, NULL, rngsymb);
	    
	    SgFunctionRefExp * funcexpr = new SgFunctionRefExp (*fname);
	    
	    SgValueExp * start= new SgValueExp(tdef->getStart(j));
	    funcexpr->addArg(*start);
	    
	    SgValueExp * end = new SgValueExp(tdef->getEnd(j));
	    funcexpr->addArg(*end);

	    SgSymbol * procname = ddef->gettosymbol();

	    // distributing code;
	    SgValueExp * dist = new SgValueExp(1);
	    funcexpr->addArg(*dist);

	    // in the following we want the grp name of the prcossor;

	    SgSymbol * procgrd = pdef->getgrp();
	    SgExpression * procgrdexpr = new SgExpression(VAR_REF, NULL, NULL, procgrd);
	    funcexpr->addArg(*procgrdexpr);
	    
	    // we need the processor dim here;

	    int processorDimCounter=0;
	    for (int k=0; k<=j; k++) {
	      if (ddef->distCode(k)!=NOP)
		processorDimCounter++;
	    }
	    SgValueExp * dim = new SgValueExp(processorDimCounter);	      

	    funcexpr->addArg(*dim);
	    
	    SgAssignStmt * assign = new SgAssignStmt(*rngarray, *funcexpr);
	    
	    lastdec->insertStmtAfter(*assign, *s);
	    break;
	  }
	default:
	  break;
	}
      }
    }

    {
      SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "delete_range", NULL);

      for (int j=tdef->getDim()-1; j>=0; j--) {
      
	SgValueExp * dimention = new SgValueExp(j+1);
	SgExpression * range = new SgExpression(RANGE_OP, NULL, dimention, NULL);
	SgExpression * ddot = new SgExpression(DDOT, range, NULL, NULL);
	SgExpression * index = new SgExpression(EXPR_LIST, ddot, NULL, NULL);

	SgExpression * rngarray = new SgExpression(ARRAY_REF, index, NULL, rngsymb);

	SgCallStmt * afuncall = new SgCallStmt(*fname);
	afuncall->addArg(*rngarray);

	lastexec->insertStmtBefore(*afuncall, *s);
      }
    }

  }
};

void add_processor(SgStatement * s, SgStatement * lastdec, SgStatement * lastexec)
{
  for (int i=0; i<s->numberOfAttributes(Proc_Def); i++) {
    Processor_def * pdef = (Processor_def *) ((s->getAttribute(i, Proc_Def))->getAttributeData());

    {
      SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "new_group_grid", NULL);

      SgFunctionRefExp * funcexpr = new SgFunctionRefExp (*fname);

      SgValueExp * dim = new SgValueExp(pdef->getDim());
      funcexpr->addArg(*dim);

      SgExpression * procshp= new SgExpression(VAR_REF, NULL, NULL, pdef->getshp());
      funcexpr->addArg(*procshp);

      SgSymbol * procgrp = pdef->getgrp();
      SgExpression * procgrpexpr = new SgExpression(VAR_REF, NULL, NULL, procgrp);
      
      SgAssignStmt * assign = new SgAssignStmt(*procgrpexpr, *funcexpr);

      lastdec->insertStmtAfter(*assign, *s);
    };
    
    for (int j =0; j< pdef->getDim(); j++) {
      SgValueExp * dimsize= new SgValueExp(pdef->getEnd(j)-pdef->getStart(j)+1);
      
      SgValueExp * dimention = new SgValueExp(j+1);
      SgExpression * range = new SgExpression(RANGE_OP, NULL, dimention, NULL);
      SgExpression * ddot = new SgExpression(DDOT, range, NULL, NULL);
      SgExpression * index = new SgExpression(EXPR_LIST, ddot, NULL, NULL);
      
      SgExpression * shparray = new SgExpression(ARRAY_REF, index, NULL, pdef->getshp());

      SgAssignStmt * assign = new SgAssignStmt(*shparray, *dimsize);
      
      lastdec->insertStmtAfter(*assign, *s);
    };
    {
      SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "delete_group", NULL);

      SgCallStmt * afuncall = new SgCallStmt(*fname);

      SgExpression *grpexp = new SgExpression(VAR_REF, NULL, NULL,pdef->getgrp());
      afuncall->addArg(*grpexp);

      lastexec->insertStmtBefore(*afuncall, *s);
    };

  }
};

void add_definition(SgStatement * s)
{
  SgStatement * lastdec=s->lastDeclaration();

  SgStatement * lastexec; // here we want the control end;
  lastexec=s->childList1(s->numberOfChildrenList1()-1);

  // add things in reverse order;
  //define_noalign_data(s, lastdec, lastexec);
  dist_align_array(s, lastdec, lastexec);
  add_template(s, lastdec, lastexec);
  add_processor(s, lastdec, lastexec);
  if (in_main_program==on)
    add_init_finalize(s, lastdec, lastexec);
  if (in_procedure==on)
    add_stack_operation(s, lastdec, lastexec);

}

//----------------;

void add_int_startwith(SgProgHedrStmt * proghedr, char * string)
{
  SgType * inttype = gettype(T_INT);

  SgSymbol * symb = proghedr->firstSymbolInScope();
  SgExprListExp * declist = new SgExprListExp();
  
  for ( ; symb; symb=symb->nextInScope()) {
    if (!strncmp(symb->identifier(), string, strlen(string))) {
      SgExpression * item = new SgExpression (VAR_REF, NULL, NULL, symb);
      declist->setLhs(*item);
      break;
    }
  }
  
  if (symb==NULL)
    return;
  else
    symb=symb->nextInScope();
  
  for ( ; symb; symb=symb->nextInScope()) {
    if (strncmp(symb->identifier(), string, strlen(string))==0) {
      SgExpression * item = new SgExpression (VAR_REF, NULL, NULL, symb);
      SgExpression * temp = declist;
      while (temp->operand(2)!=NULL)
	temp=temp->operand(2);
      SgExpression * link = new SgExpression (EXPR_LIST, item, NULL, NULL);
      temp->setRhs(*link);
    }
  }
  
  if (declist->length()!=0) {
      SgVarDeclStmt * decstmt = new SgVarDeclStmt(*declist, *inttype);
      proghedr->childList1(0)->insertStmtBefore(*decstmt, *proghedr);
  }
}

void add_new_dec(SgStatement * s)
{
  SgProgHedrStmt * proghedr = (SgProgHedrStmt *) s;

  add_include(s);

  add_int_startwith(proghedr, "pcrc_grp0");

  add_int_startwith(proghedr, "pcrc_ast");
  add_int_startwith(proghedr, "pcrc_aub");
  add_int_startwith(proghedr, "pcrc_alb");

  add_int_startwith(proghedr, "pcrc_sdx");
  add_int_startwith(proghedr, "pcrc_v");
  add_int_startwith(proghedr, "pcrc_u");

  add_int_startwith(proghedr, "pcrc_li");
  add_int_startwith(proghedr, "pcrc_irg");

  add_int_startwith(proghedr, "pcrc_tdd");
  add_int_startwith(proghedr, "pcrc_tbs");
  add_int_startwith(proghedr, "pcrc_sdd");

  add_int_startwith(proghedr, "pcrc_dad");
}



